@@ -16,9 +16,8 @@
  * Inc.,  51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
-#include <xfs/xfs.h>
-#include <xfs/command.h>
-#include <xfs/input.h>
+#include "command.h"
+#include "input.h"
 #include <sys/time.h>
 #include <sys/resource.h>
 #include "init.h"
@@ -109,8 +108,8 @@ getrusage_f(
 void
 getrusage_init(void)
 {
-	getrusage_cmd.name = _("getrusage");
-	getrusage_cmd.altname = _("g");
+	getrusage_cmd.name = "getrusage";
+	getrusage_cmd.altname = "g";
 	getrusage_cmd.argmin = 0;
 	getrusage_cmd.argmax = -1;
 	getrusage_cmd.cfunc = getrusage_f;
